"""The :mod:`lamya.site_generator` module provides a static site generator, that can
be ran either from the CLI by running `lamya` as a module or by importing
the :class:`lamya.site_generator.SiteGenerator` class and using it in your own scripts.

The static site generator utilises the :mod:`lamya.content_tree` tree implementation
to parse existing content and render it using a template engine (Jinja2 by default),
while optionally aggregating content into list pages, categories and an archive.
"""
from pathlib import Path
import glob
from datetime import datetime
from functools import partial
import os
import shutil
from collections import OrderedDict
import urllib.parse
import json

try:
    import markdown
    try:
        import markdown_strikethrough
    except ImportError:
        markdown_strikethrough = None
except ImportError:
    markdown = None
try:
    import jinja2
except ImportError:
    jinja2 = None

from lamya import content_tree
from lamya import content_processing


class Callbacks: # pylint: disable=too-few-public-methods
    """A struct for specifying callbacks to be run at potential various stages
    of the static site generation process.

    NOTE: At the moment it only contains a callback for immediately after
    :class:`lamya.content_tree.ContentTree` node being created, but I'd like to have
    an easy way of adding more callbacks in the future.
    """
    @staticmethod
    def post_content_tree_entity_create(site_generator, entity):
        """Ran on every single :class:`lamya.content_tree.ContentTree` instance created
        in the context of the site generator.

        :param site_generator: the site generator object
        :param entity: the :class:`lamya.content_tree.ContentTree` instance
        """
        # NOTE: I can use these callbacks to make sure all data that I would
        # like to exist on the content_tree entities actually does, rather
        # than defining everything on the content_tree definitions, which
        # im still not sure is a good/bad idea
        entity.site_generator_data = {
            "is_post": isinstance(entity, content_tree.PageOrPost) and\
                    not site_generator.is_page_func(entity),
            "site_url": site_generator.url}


class RenderablePage: # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """This struct is what will be passed to the template engine, so all the
    properties are seralized into directly usable information.

    :param name: the name of the page or post, as taken directly from the name
        of the file it corresponds to if it hasn't been overwritten
    :param title: falls back to the name with underscores replaced with spaces
        and capital letters at the beginning of the words if a title wasn't
        provided in the front matter
    :param content: the processed markup ready for rendering
    :param href: the relative URL of this page or post
    :param site_url: the URL of the website
    :param aggregated_posts: the list of posts if this is an aggregated page
    :param aggregated_grouped_posts: the `dict` of groups and posts if this
        is an aggregated_grouped_post
    :param pagination: the pagination `dict` if this is a paginated page
    :param publish_date: the publish date as read from the front matter or
        if one wasn't supplied, the last modified time of the source file
    :param user_data: any custom user data that has been supplied
    :param front_matter: the evaluated front matter
    :param is_post: whether ot not this page represents a post or a page
        NOTE: the heuristic for defining this is really naive by default, but the
        `is_page_func` function can be specified on the `SiteGenerator` class
    :param absolute_canonical_href: the canonical url for this page or post
    :param breadcrumbs: the breadcrumbs navigation dict
    """
    def __init__(self, page_or_post):
        """Constructor

        :param page_or_post: the page or post to create a render struct for
        """
        self.name = page_or_post.name
        self.title = page_or_post.front_matter.get(
            "title", self.name.replace("_"," ").title())
        self.content = page_or_post.content
        self.excerpt = content_processing.get_excerpt(self.content) if self.content else ""
        self.href = str(page_or_post.href)
        self.site_url = page_or_post.site_generator_data["site_url"]
        self.aggregated_posts = [] if not isinstance(page_or_post, content_tree.AggregatedPage)\
            else [self.__class__(x) for x in page_or_post.aggregated_posts]
        self.aggregated_grouped_posts = [] if not isinstance(
                page_or_post, content_tree.AggregatedGroupsPage)\
            else {k: [self.__class__(x) for x in v]\
                  for k,v in page_or_post.aggregated_grouped_posts.items()}
        self.pagination = page_or_post.pagination.as_navigation_dict()\
            if hasattr(page_or_post, "pagination")\
                and page_or_post.pagination.max_page_number > 1 else {}
        self.publish_date = page_or_post.site_generator_data.get("publish_date")
        self.user_data = page_or_post.user_data
        self.front_matter = page_or_post.front_matter
        self.is_post = page_or_post.site_generator_data.get("is_post",False)
        self.absolute_canonical_href = urllib.parse.quote_plus(self.site_url + str(
            page_or_post.site_generator_data.get("canonical_href", self.href)))
        self.breadcrumbs = [("home","/")] + [
            (p.name,str(p.href) if isinstance(p, content_tree.Folder) and p.index_page else "")\
                for p in reversed(page_or_post.ancestors[:-1])] +\
            ([(self.title if self.is_post else self.name,self.href)]\
                if (self.href != "/" and not page_or_post.is_index_page()\
                    and self.pagination.get("page_number",1) == 1) else [])


class SiteInfo: # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """This struct represents all the site information to be passed to the
    template engine.

    :param name: the name of the site
    :param url: the url of the site
    :param subtitle: the subtitle of the site, generally used as a short
        description or motto
    :param navigation: the navigation `dict`
    :param lang: the language to specify in the html `lang` property
    :param theme_options: a `dict` with the specified theme options
    :param internal_data: similar to the user data of pages, but specified
        by the site generator
    :param archive_nav: if an archive has been generated, this will be a
        navigation `dict` to all of the archive pages
    :param category_nav: if categories have been generated, this will be a
        navigation `dict` to all of the category pages
    :param display_date_format: the python `datetime` format to display
        the publish dates in
    :param author_link: a link to an author page, to be used in the `meta`
        author tag
    """
    def __init__(self, site_generator):
        """Constructor

        :param site_generator: the `SiteGenerator` object
        """
        self.name = site_generator.name
        self.url = site_generator.url
        self.subtitle = site_generator.subtitle
        self.navigation = site_generator.navigation
        self.lang = site_generator.lang
        self.theme_options = site_generator.theme_options
        self.internal_data = site_generator.internal_data
        self.archive_nav = site_generator.archive.as_navigation_dict() if\
            site_generator.archive else {}
        self.category_nav = site_generator.categories.as_navigation_dict() if\
            site_generator.categories else {}
        self.display_date_format = site_generator.display_date_format
        self.author_link = site_generator.author_link


class SiteGenerator: # pylint: disable=too-many-instance-attributes
    """This class is responsible for generating a static website by aggregating,
    grouping and rendering existing content.

    :param name: name of the site
    :param url: url of the site
    :param subtitle: the subtitle of the site, generally used as a short
        description or motto
    :param content_directory: the path to the top folder containing the site content
    :param theme_directory: the path to the theme folder
    :param static_directory: the path to a folder containing static files to be
        directly copied over into the built website
    :param templates_directory: the path to a folder with templates that will
        overwrite the theme's ones
    :param build_directory: the path to a folder where to generate the website
    :param locally_aggregate_whitelist: a list of folders that should have
        their content aggregated into an index page
    :param locally_aggregate_blacklist: a list of folders that should NOT have
        their content aggregated into an index page
    :param globally_aggregate_whitelist: a list of pages or posts to aggregate
        into the home page
    :param globally_aggregate_blacklist: a list of pages or posts to NOT aggregate
        into the home page
    :param num_posts_per_page: how many posts to have per page
    :param is_page_func: a function that returns True if a piece of content is
        meant to be treated as a static page rather than as a post
    :param front_matter_delimiter: we expect the front matter in the raw content
        of each page/post to be surrounded by lines containing only this character
    :param callbacks: a `Callbacks` struct for running functions at specific times.
        NOTE: at the moment we only run after generating any :mod:`lamya.content_tree` node
    :param lang: the language to put in the `html lang` property
    :param front_matter_publish_date_key: the name of the publish date key
        in the front matter
    :param read_date_format: the `datetime` format to expect the publish date
        in the front matter to be in
    :param display_date_format: the `datetime` format to display the publish date in
    :param author_link: a URL to be placed in the `link rel=author` html tag
    :param theme_options: a `dict` of theme options
    :param use_absolute_urls: whether or not to use full absolute URLs instead
        of the default relative ones
    :param content_tree: the :class:`lamya.content_tree.Root` node that contains the
        content of the static site
    :param internal_data: a `dict` containing extra information about the site
        generation process
    """
    def __init__(self, name, url, subtitle="", content_directory=Path("content"), # pylint: disable=too-many-arguments,too-many-locals
            theme_directory=None, static_directory=Path("static"),
            templates_directory=Path("templates"), build_directory=Path("build"),
            locally_aggregate_whitelist=None, locally_aggregate_blacklist=None,
            globally_aggregate_whitelist=None, globally_aggregate_blacklist=None,
            num_posts_per_page=-1,
            is_page_func=lambda x: isinstance(x.parent, content_tree.Root),
            front_matter_delimiter="+", callbacks=Callbacks(),
            lang="en", front_matter_publish_date_key="publish_date",
            read_date_format="%d-%m-%Y %H:%M", display_date_format="%B %-d, %Y",
            author_link="", theme_options=None, use_absolute_urls=False):
        """Constructor

        :param name: name of the site
        :param url: url of the site
        :param subtitle: optional subtitle of the site, generally used as a short
            description or motto
        :param content_directory: the path to the top folder containing the site content.
            Defaults to a folder called `content` in the current directory. This
            folder is required to exist.
        :param theme_directory: the path to the theme folder. This folder is
            required to exist. If `None` is specified, then the default theme
            will be used. `None` by default.
        :param static_directory: the path to a folder containing static files to be
            directly copied over into the built website. Defaults to a folder
            called `static` in the current directory. This folder is not
            required to exist.
        :param templates_directory: the path to a folder with templates that will
            overwrite the theme's ones. Defaults to a folder called `templates`
            in the current directory. This folder is not required to exist.
        :param build_directory: the path to a folder where to generate the website.
            Defaults to a folder called `build` in the current directory. The
            parent of this folder is required to exist.
        :param locally_aggregate_whitelist: a list of folders that should have
            their content aggregated into an index page. Defaults to []. Either
            the whitelist or the blacklist can be specified, but not both.
        :param locally_aggregate_blacklist: a list of folders that should NOT have
            their content aggregated into an index page. Defaults to []. Either
            the whitelist or the blacklist can be specified, but not both.
        :param globally_aggregate_whitelist: a list of pages or posts to aggregate
            into the home page. Defaults to []. Either the whitelist or the
            blacklist can be specified, but not both.
        :param globally_aggregate_blacklist: a list of pages or posts to NOT aggregate
            into the home page. Defaults to []. Either the whitelist or the
            blacklist can be specified, but not both.
        :param num_posts_per_page: how many posts to have per page. Defaults to
            -1, which means infinitely many posts, i.e. no pagination.
        :param is_page_func: a function that returns True if a piece of content is
            meant to be treated as a static page rather than as a post. Defaults
            to a function checking if the content's parent is the root.
        :param front_matter_delimiter: we expect the front matter in the raw content
            of each page/post to be surrounded by lines containing only this character.
            Defaults to '+'.
        :param callbacks: a `Callbacks` struct for running functions at specific times.
            Defaults to the `Callbacks` struct defined in this module.
        :param lang: the language to put in the `html lang` property. Defaults to 'en'.
        :param front_matter_publish_date_key: the name of the publish date key
            in the front matter. Defaults to `publish_date`.
        :param read_date_format: the `datetime` format to expect the publish date
            in the front matter to be in. Defaults to `%d-%m-%Y %H:%M`.
        :param display_date_format: the `datetime` format to display the publish date in.
            Defaults to `%B %-d, %Y`.
        :param author_link: a URL to be placed in the `link rel=author` html tag.
            Defaults to "".
        :param theme_options: a `dict` of theme options. Defaults to `{}`.
        :param use_absolute_urls: whether or not to use full absolute URLs instead
            of the default relative ones.
        """
        self.name = name
        self.url = url
        self.subtitle = subtitle
        self.content_directory = Path(content_directory)
        self.theme_directory = Path(theme_directory) if theme_directory is not None\
            else Path(__file__).parent / "themes" / "lamya"
        self.static_directory = Path(static_directory)
        self.templates_directory = Path(templates_directory)
        self.build_directory = Path(build_directory)
        self.locally_aggregate_whitelist = locally_aggregate_whitelist or []
        self.locally_aggregate_blacklist = locally_aggregate_blacklist or []
        self.globally_aggregate_whitelist = globally_aggregate_whitelist or []
        self.globally_aggregate_blacklist = globally_aggregate_blacklist or []
        self.num_posts_per_page = num_posts_per_page
        self.is_page_func = is_page_func
        self.front_matter_delimiter = front_matter_delimiter
        self.callbacks = callbacks
        self.lang = lang
        self.front_matter_publish_date_key = front_matter_publish_date_key
        self.read_date_format = read_date_format
        self.display_date_format = display_date_format
        self.author_link = author_link
        self.theme_options = theme_options or {}
        self.use_absolute_urls = use_absolute_urls

        if locally_aggregate_whitelist and locally_aggregate_blacklist:
            raise AggregateError("Both 'locally_aggregate_whitelist' and"
                " 'locally_aggregate_blacklist' have been specified, which is"
                " not supported, as the aggregation strategies is chosen depending"
                " on which one is specified, and if neither is, aggregation is"
                " enabled on everything by default.")

        self.globally_aggregated_posts = None
        self.categories = None
        self.archive = None
        self.navigation = None

        self.content_tree  = content_tree.ContentTree.from_directory(content_directory,
            post_create_callback=partial(callbacks.post_content_tree_entity_create, self))
        self.initialize_renderer()
        self.initialize_markup_processor()

        self.internal_data = {
            "build_date": datetime.now()
        }

    def initialize_renderer(self):
        """This method initializes the template engine which will render the site.

        By default we use `jinja2`, but overwriting this function and the render
        fundtion, should allow other template engines to be used."""
        if jinja2 is None:
            raise NotImplementedError("jinja2 was not found, so it either"
                " needs to be installed or you need to overwrite the"
                " 'initialize_renderer' method on the 'Site' class to use"
                " your own templating engine.")

        self.renderer = Jinja2Renderer([
            str(self.templates_directory),
            str(self.theme_directory / "templates")])

        def pyg_urlencode(url):
            jinja_urlencode = self.renderer.environment.filters["urlencode"]
            if self.use_absolute_urls and url.startswith("/"):
                if not self.url.startswith("file://") or "." in url.split("/")[-1]:
                    return self.url +\
                        jinja_urlencode(url)
                else:
                    return self.url +\
                        jinja_urlencode(url + "/index.html")

            parsed_url = urllib.parse.urlparse(url, allow_fragments=False)

            return parsed_url._replace(
                path=jinja_urlencode(parsed_url.path),
                params=jinja_urlencode(parsed_url.params),
                query=jinja_urlencode(parsed_url.query)).geturl()

        self.renderer.environment.filters["pyg_urlencode"] = pyg_urlencode

    def initialize_markup_processor(self):
        """This method initializes the function which we will use to process
        the markup into `html`.

        By default we use `markdown`, but overwriting this, should allow other
        markup processors to be used.
        """
        if markdown is None:
            raise NotImplementedError("SiteGenerator requires a"
                " 'markup_processor_func' to be initialized , which is set to"
                " markdown by default, but markdown is not included in the"
                " requirements list, so you need to install it separately.")

        self.markup_processor_func = lambda x:\
            markdown.markdown(x,
                extensions=[
                    "footnotes","tables","fenced_code","toc","codehilite"] +\
                    ([markdown_strikethrough.StrikethroughExtension()]\
                        if markdown_strikethrough else []),
                extension_configs = {
                    "codehilite" : {
                        "linenums": False
                    }
                })

    @staticmethod
    def run_from_config(config, render=True):
        """Use a :class:`lamya.config.Config` object to configure the whole
        generation process, which is all contained in this one function.

        Notably, the only reason to use this instead of providing a file with
        all the parameters to the `--from_file` argument of the `CLI`, is
        to have some other optional custom steps between building the site
        and rendering it.

        :param config: the :class:`lamya.config.Config` object, containing all the arguments
        :param render: whether or not to render the website at the end of the
            function or leave that to the user, so other changes can be made before
            that happens
        :returns: the built :class:`SiteGenerator` object
        """
        site_gen = SiteGenerator(
            name=config.name, url=config.url, subtitle=config.subtitle,
            content_directory=config.content_directory,
            theme_directory=config.theme_directory,
            templates_directory=config.templates_directory,
            static_directory=config.static_directory,
            build_directory=config.build_directory,
            locally_aggregate_whitelist=config.locally_aggregate_whitelist,
            locally_aggregate_blacklist=config.locally_aggregate_blacklist,
            globally_aggregate_whitelist=config.globally_aggregate_whitelist,
            globally_aggregate_blacklist=config.globally_aggregate_blacklist,
            num_posts_per_page=config.posts_per_page, lang=config.language,
            read_date_format=config.read_date_format,
            front_matter_publish_date_key=config.publish_date_key,
            display_date_format=config.display_date_format, author_link=config.author_link,
            theme_options=config.theme_options, use_absolute_urls=config.use_absolute_urls)
        site_gen.process_content_tree()
        site_gen.aggregate_posts()

        if config.build_categories:
            site_gen.build_category_pages(
                allow_uncategorized=not config.do_not_allow_uncategorized,
                uncategorized_name=config.uncategorized_name,
                category_list_page_name=config.categories_page_name,
                group=config.group_categories)

        if config.build_archive_by_month or config.build_archive_by_year:
            site_gen.build_archive(
                config.archive_month_format, config.archive_year_format)
            site_gen.build_archive_pages(
                by_month=config.build_archive_by_month, by_year=config.build_archive_by_year,
                archive_list_page_name=config.archive_page_name, group=config.group_archive,
                display_by_month_in_list_page=config.display_archive_by_month_in_list_page,
                display_by_year_in_list_page=config.display_archive_by_year_in_list_page)

        if config.custom_navigation is None:
            site_gen.build_navigation(
                extra_filter_func=lambda x: x.name not in config.exclude_from_navigation\
                                        and str(x.path) not in config.exclude_from_navigation,
                exclude_categories=config.exclude_categories_from_navigation,
                exclude_archive=config.exclude_archive_from_navigation)

            if config.home_name_in_navigation is not None:
                site_gen.navigation.update({config.home_name_in_navigation: "/"})
                site_gen.navigation.move_to_end(config.home_name_in_navigation, last=False)
        else:
            def recursive_parse_paths(_dict):
                return {
                    k: (recursive_parse_paths(v) if isinstance(v, dict) else v)\
                    for k,v in _dict.items()}

            site_gen.navigation = recursive_parse_paths(json.loads(config.custom_navigation)\
                if isinstance(config.custom_navigation, str) else config.custom_navigation)

        if render:
            site_gen.render()

        return site_gen

    def process_content_tree(self):
        """This method goes through all of the content, read from the `content`
        directory and ensures we have all the data we require for them.

        At the moment it only tries to read a `publish_date` for each page.
        """
        ## At this point we have the main hierarchy. Let's now read the source, so
        # we can start optionally grouping the content using different heuristics
        for page in self.content_tree.leaves():
            page.parse_front_matter_and_content()

            # We will be sorting content by date, so let's make sure all content
            # has some sort of a date, either explicit in the front matter or
            # we take the last modified time as a backup
            if page.site_generator_data.setdefault("publish_date", None) is None:
                # Use an `if`, so we can support the user manually setting
                # the publish date beforehand if need be
                last_modified_time = datetime.fromtimestamp(
                    os.path.getmtime(page.source_path) if page.source_path else 0)

                front_matter_publish_date = getattr(page, "front_matter", {}).get(
                    self.front_matter_publish_date_key)
                front_matter_publish_date = datetime.strptime(
                    front_matter_publish_date, self.read_date_format) if\
                    front_matter_publish_date else None

                page.site_generator_data["publish_date"] =\
                    front_matter_publish_date or last_modified_time

                if not page.site_generator_data["publish_date"]:
                    content_tree.warning(f"There's no '{self.front_matter_publish_date_key}'"
                        " key in the front matter for '{page.path}' and neither"
                        " is there a 'source_path' that we can read the last "
                        " modified time from, so the date is going to be 0")

    def aggregate_posts(self):
        """This method uses the aggregation arguments passed to the constructor
        to aggregate posts into index pages.
        """
        ## Aggregate folders with no index pages
        # Check if any folder is missing an index page, which would mean
        # we need to create one. NOTE: most of the time I would imagine this
        # would be the case, as the way I understand using folders is to
        # separate different types of content - blog, archive, portfolio, etc.
        folders_with_no_index = list(filter(
            lambda x: isinstance(x, content_tree.Folder) and x.index_page is None,
            self.content_tree.flat(include_index_pages=False)))

        # Depending on the aggregate whitelists we have different strategies
        to_locally_aggregate = folders_with_no_index

        if self.locally_aggregate_blacklist:
            to_locally_aggregate = [x for x in folders_with_no_index\
                if x.name not in self.locally_aggregate_blacklist and\
                   str(x.path) not in self.locally_aggregate_blacklist]

        if self.locally_aggregate_whitelist:
            to_locally_aggregate = [x for x in folders_with_no_index\
                if x.name in self.locally_aggregate_whitelist or\
                   str(x.path) in self.locally_aggregate_whitelist]

        for folder in to_locally_aggregate:
            folder.index_page = content_tree.AggregatedPage(
                folder.name, sorted(
                    [x for x in folder.children if isinstance(x,content_tree.PageOrPost)],
                    reverse=True, key=lambda x: x.site_generator_data["publish_date"]))
            self.callbacks.post_content_tree_entity_create(self, folder.index_page)
            if self.num_posts_per_page > 0:
                folder.index_page.paginate(self.num_posts_per_page,
                    partial(self.callbacks.post_content_tree_entity_create, self))

        ## Aggregate all posts to optionally be used on the home page
        to_globally_aggregate = list(filter(
            lambda x: not self.is_page_func(x) and\
                      not isinstance(x, content_tree.PaginatedAggregatedPage),
            self.content_tree.leaves(include_index_pages=False)))

        if self.globally_aggregate_blacklist:
            to_globally_aggregate = [x for x in to_globally_aggregate\
                if x.parent and x.parent.name not in self.globally_aggregate_blacklist and\
                   str(x.parent.path) not in self.globally_aggregate_blacklist]

        if self.globally_aggregate_whitelist:
            to_globally_aggregate = [x for x in to_globally_aggregate\
                if x.parent and (x.parent.name in self.globally_aggregate_whitelist or\
                   str(x.parent.path)) in self.globally_aggregate_whitelist]

        # Check if we have a home index page in which case we'll just store
        # the globally aggregated content and if not we'll create an AggregatedPage
        self.globally_aggregated_posts = to_globally_aggregate
        if not self.content_tree.index_page:
            self.content_tree.index_page = content_tree.AggregatedPage(
                "home", sorted(to_globally_aggregate, reverse=True,
                    key=lambda x: x.site_generator_data["publish_date"]))
            self.callbacks.post_content_tree_entity_create(
                    self, self.content_tree.index_page)
            if self.num_posts_per_page > 0:
                self.content_tree.index_page.paginate(self.num_posts_per_page,
                    partial(self.callbacks.post_content_tree_entity_create, self))

    def build_category_pages(self, parent=None, # pylint: disable=too-many-arguments,too-many-locals
            category_accessor= lambda x: getattr(x,"front_matter",{}).get("category",""),
            allow_uncategorized=True, uncategorized_name="Uncategorized",
            leaves_filter= lambda x: True,
            category_list_page_name="categories", group=True):
        """This method reads categories from the content's front matter and
        groups them into category pages.

        :param parent: the :mod:`lamya.content_tree` parent node to put the
            category pages in. If `None` the root will be used.
        :param category_accessor: a function defining how to query the category
            for each page or post node. By default it attempts to read it from
            the `category` key in the front matter.
        :param allow_uncategorized: whether or not to allow some pages or posts
            to be uncategorized or to error out if such pages or posts are found
        :param uncategorized_name: the name of the category to use for posts
            that do not have a category
        :param leaves_filter: an optional function to be used for filtering
            the leaves that should be processed
        :param category_list_page_name: the name to give to the page, containing
            all categories and their posts. If it is falsy, no such page will
            be generated.
        :param group: whether or not to group all category pages under a
            `categories` folder
        """
        grouped = {}
        for p in filter(lambda x: isinstance(x, content_tree.PageOrPost)\
                              and leaves_filter(x)\
                              and not self.is_page_func(x),
                self.content_tree.leaves()):
            grouped.setdefault(category_accessor(p), []).append(p)

        if "" in grouped:
            if not allow_uncategorized:
                raise UncategorizedNotAllowedError("The following pages don't have"
                    " categories, but 'allow_uncategorized' is False " + str(grouped[""]))
            grouped[uncategorized_name] = grouped.pop("")

        for category in grouped:
            grouped[category] = sorted(grouped[category],
                key=lambda x: x.site_generator_data["publish_date"], reverse=True)

        parent = parent or self.content_tree
        category_pages = {}
        all_category_pages = []
        for category, pages in grouped.items():
            aggregated_page = content_tree.AggregatedPage(category or uncategorized_name, pages)
            self.callbacks.post_content_tree_entity_create(self, aggregated_page)
            parent.add_child(aggregated_page)

            if self.num_posts_per_page > 0:
                paginated_pages = aggregated_page.paginate(self.num_posts_per_page,
                    partial(self.callbacks.post_content_tree_entity_create, self))
                category_pages[category] = paginated_pages[0]
                all_category_pages += paginated_pages
            else:
                all_category_pages.append(aggregated_page)
                category_pages[category] = aggregated_page

        categories_folder = None if not group else\
            parent.group((category_list_page_name or "categories"), all_category_pages)

        category_list_page = None
        if category_list_page_name and category_pages:
            category_list_page = content_tree.AggregatedGroupsPage(
                category_list_page_name, grouped)
            self.callbacks.post_content_tree_entity_create(self, category_list_page)
            if categories_folder:
                categories_folder.index_page = category_list_page
            else:
                parent.add_child(category_list_page)

        self.categories = Categories(grouped, category_pages,
            category_list_page, categories_folder, uncategorized_name)

    def build_archive(self, by_month_format="%B, %Y", by_year_format="%Y"):
        """This method builds the `archive` struct, with all the posts in
        their correct groups.

        :param by_month_format: the `datetime` format to use for grouping and
            sorting posts by month
        :param by_year_format: the `datetime` format to use for grouping and
            sorting posts by year

        :returns: the built `Archive` struct
        """
        by_month, by_year = {}, {}
        for post in self.posts():
            by_month.setdefault(post.site_generator_data["publish_date"]\
                .strftime(by_month_format),[]).append(post)
            by_year.setdefault(post.site_generator_data["publish_date"]\
                .strftime(by_year_format),[]).append(post)

        self.archive = Archive(
            OrderedDict(sorted(by_month.items(), reverse=1,
                key=lambda x: datetime.strptime(x[0], by_month_format))),
            OrderedDict(sorted(by_year.items(), reverse=1,
                key=lambda x: datetime.strptime(x[0], by_year_format))))

        return self.archive

    def build_archive_pages(self, parent=None, # pylint: disable=too-many-arguments,too-many-locals
            by_month=True, by_year=False,
            archive_list_page_name="archive", group=True,
            display_by_month_in_list_page=True,
            display_by_year_in_list_page=True):
        """This method takes the previously built `Archive` struct and generates
        the archive pages.

        :param parent: the :mod:`lamya.content_tree` parent node to put the
            archive pages in
        :param by_month: whether or not to write pages for the monthly archive
        :param by_year: whether or not to write pages for the yearly archive
        :param archive_list_page_name: the name to give to the page, containing
            all archive groups and their posts. If false, no such page will be generated.
        :param group: whether or not to group all archive pages under an
            `archives` folder
        :param display_by_month_in_list_page: whether to include the monthly
            archive in the list page
        :param display_by_year_in_list_page: whether to include the yearly archive
            in the list page
        """
        if not hasattr(self, "archive"):
            raise ArchivePagesError(
                "The archive must first be initialized using 'build_archive'"
                " before pages can be made out of it.")

        parent = parent or self.content_tree
        archives = (
            (list(self.archive.posts_by_month.items()) if by_month else []),
            (list(self.archive.posts_by_year.items()) if by_year else []))

        if not (archives[0] or archives[1]):
            raise ArchivePagesError("No posts to put in archive")

        archive_pages = []
        for archive in archives:
            archive_pages.append([])
            for archive_key, posts in archive:
                aggregated_page = content_tree.AggregatedPage(archive_key, posts)
                self.callbacks.post_content_tree_entity_create(self, aggregated_page)
                parent.add_child(aggregated_page)

                if self.num_posts_per_page > 0:
                    archive_pages[-1] +=\
                        aggregated_page.paginate(self.num_posts_per_page,
                            partial(self.callbacks.post_content_tree_entity_create, self))
                else:
                    archive_pages[-1].append(aggregated_page)

        archive_folder = None if not group else\
            parent.group(archive_list_page_name or "archive",
                    archive_pages[0] + archive_pages[1])

        archive_list_page = None
        if archive_list_page_name and (archive_pages[0] or archive_pages[1]) and\
                (display_by_month_in_list_page or display_by_year_in_list_page):
            archive_list = OrderedDict(
                self.archive.posts_by_month if display_by_month_in_list_page else {})
            archive_list.update(
                self.archive.posts_by_year if display_by_year_in_list_page else {})

            archive_list_page = content_tree.AggregatedGroupsPage(
                archive_list_page_name, archive_list)
            self.callbacks.post_content_tree_entity_create(self, archive_list_page)
            if archive_folder:
                archive_folder.index_page = archive_list_page
            else:
                parent.add_child(archive_list_page)

        self.archive.pages_by_month = archive_pages[0]
        self.archive.pages_by_year = archive_pages[1]
        self.archive.list_page = archive_list_page

    def posts(self):
        """Returns all leaves that DO NOT pass the `is_page_func` test."""
        return [x for x in self.content_tree.leaves() if\
                isinstance(x, content_tree.PageOrPost) and not self.is_page_func(x)]

    def build_navigation(self, filter_func=None, extra_filter_func=None,
            exclude_categories=False, exclude_archive=False):
        """Builds a navigation structure.

        :param filter_func: a function to filter out pages from the navigation.
            If `None` a default function will be used, which boils down to adding
            pages to the navigation if any of the following are true:

            - the node passes the `is_page_func` or it is a
                :class:`lamya.content_tree.Folder`
            - the node is an `index_page`
            - the node is a `category` page and categories are not excluded
            - the node is an `archive` page and the archive is not excluded

            AND BOTH of the following are true:

            - this node is not the home page
            - if this node is part of a pagination, it is the first page, otherwise
                assume `True`
        :param extra_filter_func: a second filtering function, so if a bit of
            extra filtering is required on top of the default function it can
            easily be provided
        :param exclude_categories: whether or not to exclude any category pages
            from the navigation
        :param exclude_archive: whether or not to exclude any archive pages
            from the navigation
        """
        if self.navigation:
            content_tree.warning("Navigation already exists, overwriting..")

        extra_filter_func = extra_filter_func or (lambda _: True)
        is_category_page_func = lambda x: False if not self.categories\
            else x.path in [p.path for p in list(self.categories.all_pages) +\
                ([self.categories.list_page] if self.categories.list_page else [])]
        is_archive_page_func = lambda x: False if not self.archive\
            else x.path in [p.path for p in self.archive.all_pages +\
                ([self.archive.list_page] if self.archive.list_page else [])]

        filter_func = filter_func or (lambda x:\
               (((isinstance(x, content_tree.AbstractPageOrPost) and self.is_page_func(x))\
                    and not is_category_page_func(x) and not is_archive_page_func(x))\
                or isinstance(x, content_tree.Folder)\
                or (isinstance(x, content_tree.AbstractPageOrPost) and x.is_index_page())\
                or (not exclude_categories and is_category_page_func(x))\
                or (not exclude_archive and is_archive_page_func(x)))\
            and x != self.content_tree.index_page and\
            (x.pagination.page_number == 1 if hasattr(x, "pagination") else True))

        navigatable_tree = self.content_tree.filter(
            lambda x: (filter_func(x) if filter_func else True)\
                      and (extra_filter_func(x) if extra_filter_func else True), True)
        self.navigation = navigatable_tree.as_dict(
            lambda x: str(x.href),
            lambda x: str(x.href)\
                if isinstance(x, content_tree.Folder) and x.index_page else None)

    def render(self,
            to_renderable_page=RenderablePage, to_site_info=SiteInfo,
            markup_processor_func=None,
            template_accessor_func=lambda x: x.front_matter.get("template","default.html"),
            **kwargs):
        """This method renders the site into the build folder.

        :param to_renderable_page: a function that converts an `AbstractPageOrPost`
            into a struct containing the information necessary for rendering that
            page or post. That struct is then passed to the template engine.
            Default is `RenderablePage`.
        :param to_site_info: a function that converts the `SiteGenerator`
            into a struct containing the information necessary for the template.
            Default is `SiteInfo`.
        :param markup_processor_func: an optional markup processing function.
            If `None` the `markup_processor_func` property of the class will
            be used. Default is `None`.
        :param template_accessor_func: a function specifying how to choose a
            template for each leaf node. Defaults to checking if there's a
            'template' argument in the front matter and if not uses 'default.html'.
        :param kwargs: any other optional information to pass to the template engine
        """
        if markup_processor_func is None:
            markup_processor_func = self.markup_processor_func

        if self.build_directory.exists():
            shutil.rmtree(self.build_directory)

        self.build_directory.mkdir()

        # Copy over static content
        def copy_static(static_dir):
            if static_dir.exists():
                for x in glob.glob(str(static_dir) + "/**/*", recursive=True):
                    p = Path(x)
                    if not p.is_file():
                        continue

                    destination = self.build_directory / p.relative_to(static_dir)
                    if not destination.parent.exists():
                        os.makedirs(destination.parent)

                    shutil.copy2(x,destination)

        copy_static(self.theme_directory / "static")
        copy_static(self.static_directory)

        # We want to render aggregated pages last, so we make sure that all
        # posts that they aggregate already have their content processed
        for leaf in sorted(self.content_tree.leaves(),
                key=lambda x: len(getattr(x, "aggregated_posts", []) +\
                                  list(getattr(x, "aggregated_grouped_posts", {}).keys()))):
            # Make sure the content has been read and processed
            if leaf._content is None or leaf._front_matter is None:
                leaf.parse_front_matter_and_content()
                leaf.process_content(markup_processor_func)

            # We may need to write the same page to multiple paths, because
            # of pagination, so we store the path in a list
            paths = [leaf.render_path(self.build_directory)]

            # Let's write the first page of each pagination both to the
            # respective '/' and '/page1' URLs, as it's a bit awkward otherwise.
            if isinstance(leaf, content_tree.PaginatedAggregatedPage)\
                    and leaf.pagination.page_number == 1\
                    and leaf.pagination.max_page_number != 1:
                # If it's already the index page, let's also write it to /page1
                if leaf == getattr(leaf.parent, "index_page", None):
                    paths.append(paths[0].parent / "page1" / "index.html")
                else:
                    # Otherwise, we need to write to the index page
                    paths.append(paths[0].parent.parent / "index.html")
                    # Make sure we consistently use '/' as the canonical URL
                    leaf.site_generator_data["canonical_href"] =\
                        (leaf.parent.href / leaf.name) if leaf.parent else ""

            for path in paths:
                if not path.parent.exists():
                    os.makedirs(path.parent)

                with open(path, "w", encoding="utf-8") as f:
                    f.write(self.renderer.render(template_accessor_func(leaf),
                        page=to_renderable_page(leaf),
                        site_info=to_site_info(self), **kwargs))

        # Make sure we don't have folders with no index files in them, which
        # would show a file browser in the web browser
        class _404Page: # pylint: disable=too-few-public-methods
            front_matter = {}
        for folder in filter(lambda x: isinstance(x, content_tree.Folder),
                self.content_tree.flat(False) + [self.content_tree]):
            if not folder.index_page:
                if not folder.leaves():
                    continue
                folder_render_path = self.build_directory /\
                    folder.path.relative_to("/") / "index.html"
                if not folder_render_path.parent.exists():
                    os.makedirs(folder_render_path.parent)

                with open(folder_render_path, "w",
                        encoding="utf-8") as f:
                    f.write(self.renderer.render("404.html",
                        page=_404Page(),
                        site_info=to_site_info(self)))

                if folder == self.content_tree:
                    content_tree.warning(
                        "There's no home index page. Consider writing a custom"
                        " one or calling the `SiteGenerator.aggregate_posts method`.")


class Jinja2Renderer: # pylint: disable=too-few-public-methods
    """A thin wrapper around the jinja2 environment to provide a framework
    for using different renderers.

    :param environment: the `jinja2` environment
    """
    def __init__(self, template_directories):
        """Constructor

        :param template_directories: a number of paths to look for templates in
        """
        self.environment = jinja2.Environment(
            loader=jinja2.ChoiceLoader(
                [jinja2.FileSystemLoader(x) for x in template_directories]),
            autoescape=jinja2.select_autoescape(),
            trim_blocks=True, lstrip_blocks=True)

    def render(self, template, **render_data):
        """Renders the given template, with the passed in render date.

        :param template: the `template` to use for rendering
        :param render_data: any data to be passed to the template engine
        """
        return self.environment.get_template(template).render(**render_data)


class Categories: # pylint: disable=too-many-arguments
    """A struct grouping all the relevant categories information.

    :param posts_by_category: a `dict` containing all the posts belonging
        to specific categories
    :param pages_by_category: a `dict` mapping a category name to the category page
    :param list_page: the optional page containing all categories and their posts
        that may have been built
    :param folder: an optional folder if the categories have been grouped
    :param uncategorized_name: the name of the uncategorized category
    """
    def __init__(self, posts_by_category, pages_by_category, list_page, folder,
            uncategorized_name):
        """Constructor

        :param posts_by_category: a `dict` containing all the posts belonging
            to specific categories
        :param pages_by_category: a `dict` mapping a category name to the category page
        :param list_page: the optional page containing all categories and their posts
            that may have been built
        :param folder: an optional folder if the categories have been grouped
        :param uncategorized_name: the name of the uncategorized category
        """
        self.posts_by_category = posts_by_category
        self.pages_by_category = pages_by_category
        self.list_page = list_page
        self.folder = folder
        self.uncategorized_name = uncategorized_name

    @property
    def all_pages(self):
        """This method returns all category pages, except for the list page."""
        return self.pages_by_category.values()

    def as_navigation_dict(self):
        """This method maps the categories structure into a simple `dict`,
        so it can be accessed in the template engine.
        """
        return {
            "self": getattr(self.list_page, "href", None),
            "categories": {k: str(v.href) for k,v in self.pages_by_category.items()},
            "uncategorized_name": self.uncategorized_name }


class Archive:
    """A struct grouping all of the relevant archives information.

    :param posts_by_month: the posts grouped into a mapping by month
    :param posts_by_year: the posts grouped into a mapping by year
    :param pages_by_month: the pages grouped into a mapping by month
    :param pages_by_year: the pages grouped into a mapping by year
    :param list_page: the optional page containing all archives and their posts
    """
    def __init__(self, posts_by_month, posts_by_year):
        """Constructor

        :param posts_by_month: the posts grouped into a mapping by month
        :param posts_by_year: the posts grouped into a mapping by year
        :param pages_by_month: the pages grouped into a mapping by month
        :param pages_by_year: the pages grouped into a mapping by year
        :param list_page: the optional page containing all archives and their posts
        """
        self.posts_by_month = posts_by_month
        self.posts_by_year = posts_by_year
        self.pages_by_month = []
        self.pages_by_year = []
        self.list_page = None

    @property
    def all_pages(self):
        """This property returns all pages by month and year."""
        return self.pages_by_month + self.pages_by_year

    def as_navigation_dict(self):
        """This method returns a simple `dict` containing all the archive posts
        and pages information, so it can be easily accessed in the template engine.
        """
        return {
            "by_month": [(p.name, str(p.href)) for p in self.pages_by_month\
                if not hasattr(p, "pagination") or p.pagination.page_number == 1],
            "by_year": [(p.name, str(p.href)) for p in self.pages_by_year\
                if not hasattr(p, "pagination") or p.pagination.page_number == 1]
            }


class AggregateError(Exception):
    """Raised when attempting to use both a whitelist and a blacklist for
    specifying content to be aggregated."""

class UncategorizedNotAllowedError(Exception):
    """Raised when a category is required, but none has been specified."""

class ArchivePagesError(Exception):
    """Raised when an attempt is made to create the archive pages, before the
    content has been created or when there's nothing to archive."""



"""The lamya (emphasis is on the ya - lamyà) package provides both an out of
the box markdown static site generator and a framework for building your own
custom ones.
"""
from .content_tree import ContentTree, Folder, Root, AbstractPageOrPost,\
    PageOrPost, ProceduralPage, AggregatedPage, PaginatedAggregatedPage,\
    AggregatedGroupsPage



"This module provides a few utility functions for processing content and front matter."
import re


def split_front_matter(source, front_matter_delimiter="+"):
    """Splits and evaluates the front matter from the content's source.

    This function assumes the front matter is actual python code, which gets
    `exec`-ed to give us a dictionary of values.

    :param source: the input raw source
    :param front_matter_delimiter: the delimiting character used for separating
        front matter from the rest of the content. The front matter should be
        surrounded by lines containing only delimiter characters.
    """
    if not source:
        return {}, ""

    as_lines = source.splitlines()
    has_front_matter = set(as_lines[0]) == set([front_matter_delimiter])

    if not has_front_matter:
        return {}, source

    front_matter = ""
    content = ""
    for i, line in enumerate(as_lines[1:]):
        if set(line) == set(front_matter_delimiter):
            content = "\n".join(as_lines[i+2:])
            break
        front_matter += line + "\n"

    evaluated_front_matter = {}
    exec(front_matter, {}, evaluated_front_matter) # pylint: disable=exec-used

    return evaluated_front_matter, content


def get_excerpt(content, remove_html_tags=True, # pylint: disable=too-many-arguments
        excerpt_start_tag="<!--excerpt-start-->",
        excerpt_end_tag="<!--excerpt-end-->",
        fallback_num_characters=250, suffix=""):
    """Extracts an excerpt from the provided content.

    The excerpt can be extracted either from excerpt tags if they exist in
    the content or by just taking the first ~150 characters of clean text,
    where clean means without any html tags.

    :param content: the content to grab the excerpt from. Can either be processed
        or raw markup
    :param remove_html_tags: whether or not to try removing html tags from
        the excerpt
    :param excerpt_start_tag: an optional piece of string that can be use for
        specifying the beginning of the excerpt
    :param excerpt_end_tag: an optional piece of string that can be use for
        specifying the ending of the excerpt
    :param fallback_num_characters: if no delimiting tags were specified,
        take everything up to the first space after `fallback_num_characters`
        as the excerpt
    :param suffix: optional suffix to add at the end of the excerpt. For when
        something like `...` is appropriate
    """
    if excerpt_start_tag in content or excerpt_end_tag in content:
        excerpt = str(content)
        if excerpt_start_tag in content:
            excerpt = excerpt.split(excerpt_start_tag, 1)[1]
        if excerpt_end_tag in content:
            excerpt = excerpt.split(excerpt_end_tag, 1)[0]
        excerpt = remove_html(excerpt) if remove_html_tags else excerpt
    else:
        clean_content = remove_html(content)
        i = 0
        for i, char in enumerate(clean_content[fallback_num_characters:]):
            if char == " ":
                break
        excerpt = str(clean_content[:fallback_num_characters+i])
    return "".join(excerpt.splitlines()) + suffix


def remove_html(x):
    """Uses regex to remove html tags.

    :param x: the string to try to rid of html
    """
    # https://stackoverflow.com/questions/9662346/python-code-to-remove-html-tags-from-a-string
    clean_re = re.compile('<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});')
    return re.sub(clean_re, "", x).replace("\n"," ")



import argparse
import os
import sys
from pathlib import Path
import json

from lamya import site_generator


class UnrecognizedLamyaArgumentError(Exception):
    pass

class NotEnoughValuesForLamyaArgumentError(Exception):
    pass

class MissingRequiredLamyaArgumentError(Exception):
    pass


def build_parser():
    parser = argparse.ArgumentParser(
        prog="python -m lamya",
        description="An opinionated static site generator using the `lamya` library",
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("-ff","--from_file",
        help="read the arguments from a python file, e.g. "
            "`python -m lamya -ff config.py`. If supplied, all extra command line"
            " arguments will be treated as overrides for the config file.")
    parser.add_argument("-n","--name",
        help="the name of the website. Defaults to current directory name.")
    parser.add_argument("-url","--url")
    parser.add_argument("-st","--subtitle",
        help="the subtitle of the website, usually a short sentence. Defaults to ''.")
    parser.add_argument("-al","--author_link", default="",
        help="author link to be used as a fallback for page/post front matter"
            " author_link key, when specifying the author using the"
            " <link rel='author' ..> tag. Defaults to the ''.")
    parser.add_argument("-l","--language", default="en",
        help="the language of the website as required by the HTML `lang` attribute. "
        "Default is 'en'.")
    parser.add_argument("-uau","--use_absolute_urls",
        action="store_const", const=True, default=False,
        help="whether or not to use absolute URLs. It is only recommended to use"
            " this flag if you'd like to access the website directly from the"
            " filesystem using the 'file://' scheme, rather than serving it.")
    parser.add_argument("-pp","--posts_per_page", default=-1, type=int,
        help="number of posts on per page (pagination). "
            "If set to -1 no pagination will be applied.")
    parser.add_argument("-pdk","--publish_date_key", default="publish_date",
        help="the key used to specify the 'publish_date' in the front matter."
            " Defaults to 'publish_date'.")
    parser.add_argument("-rdf","--read_date_format", default="%d-%m-%Y %H:%M",
        help="the expected date format for reading the 'publish_date' key in the"
            " front matter. Default is %%d-%%m-%%Y %%H:%%M.")
    parser.add_argument("-ddf","--display_date_format", default="%B %-d, %Y",
        help="the format to display the publish date in. Default is %%B %%-d, %%Y.")

    paths_group = parser.add_argument_group("Paths")
    paths_group.add_argument("-sd","--site_directory", default=".",
        help="the path to the site directory. Defaults to the current directory.")
    paths_group.add_argument("-cd","--content_directory",
        help="the path to the content directory of the site."
            " Defaults to `{site_directory}/content`.")
    paths_group.add_argument("-thd","--theme_directory",
        help="the path to the directory of the theme."
            " Defaults to `{site_directory}/theme`.")
    paths_group.add_argument("-td","--templates_directory",
        help="the path to the a directory, containing optional overrides of the"
            " theme's templates. Defaults to `{site_directory}/templates`.")
    paths_group.add_argument("-std","--static_directory",
        help="the path to a directory, containing static files that will directly"
            " be copied over to the `build` directory."
            " Defaults to `{site_directory}/static`.")
    paths_group.add_argument("-bd","--build_directory",
        help="the directory to build the site in."
            " Defaults to `{site_directory}/build`.")

    navigation_group = parser.add_argument_group("Navigation")
    navigation_group.add_argument("-hn", "--home_name_in_navigation",
        help="the name under which the `home` page should "
            "appear in the navigation. `None` will exclude it. Default is `None`.")
    navigation_group.add_argument("-ecn", "--exclude_categories_from_navigation",
        default=False, action="store_const", const=True,
        help="whether to exclude category pages from navigation. Default is False.")
    navigation_group.add_argument("-ean", "--exclude_archive_from_navigation",
        default=False, action="store_const", const=True,
        help="whether to exclude archive pages from navigation. Default is False.")
    navigation_group.add_argument("-en", "--exclude_from_navigation",
        default=[], action="extend", nargs="+", type=str,
        help="pages to exclude from navigation, given as a list of "
            "names and/or paths, e.g. ['about','blog/personal']. Default is [].")
    navigation_group.add_argument("-cn", "--custom_navigation",
        help="accepts a `json` dictionary in the format "
            "{'name1': '/url1', 'name2': {'name2_sub1': '/name2/url_sub2'}, .. }."
            " Default is `None`. If specified will override any automated navigation generation.")

    aggregation_group = parser.add_argument_group("Aggregation",
        "By default, if any folders in the `content_directory` don't have a "
        "defined `index` page, will have an automated one created, which will "
        "be an aggregate of all the posts inside of it (local aggregation)."
        " Additionally, if there's no top-level `index` page defined, it "
        "will be an aggregate of all the posts in the website "
        "(global aggregation). The following options control how both those "
        "aggregators work.")
    aggregation_group.add_argument("-law", "--locally_aggregate_whitelist",
        default=[], action="extend", nargs="+", type=str,
        help="specifies the folder paths or names to aggregate. If nothing is"
            " provided it defaults to aggregating everything.")
    aggregation_group.add_argument("-lab", "--locally_aggregate_blacklist",
        default=[], action="extend", nargs="+", type=str,
        help="specifies the folder paths or names to NOT aggregate. If nothing is"
            " provided it defaults to nothing. NOTE: only the `whitelist` or"
            " the `blacklist` being set is supported, but not both.")
    aggregation_group.add_argument("-gaw", "--globally_aggregate_whitelist",
        default=[], action="extend", nargs="+", type=str,
        help="specifies the folder paths or names to aggregate into the home page."
            " If nothing is provided it defaults to aggregating everything.")
    aggregation_group.add_argument("-gab", "--globally_aggregate_blacklist",
        default=[], action="extend", nargs="+", type=str,
        help="specifies the folder paths or names to NOT aggregate into the home page"
            ". If nothing is provided it defaults to nothing. NOTE: only the "
            "`whitelist` or the `blacklist` being set is supported, but not both.")

    category_group = parser.add_argument_group("Category pages",
        "Building an aggregation of posts per category defined in their front matter.")
    category_group.add_argument("-bc","--build_categories",
        action="store_const", const=True, default=False,
        help="whether or not to build category pages, based on the 'category'"
            " key of their 'front_matter'. Defaults to `False`.")
    category_group.add_argument("-ac","--do_not_allow_uncategorized",
        action="store_const", const=True, default=False,
        help="whether to error if there are uncategorized posts. Defaults to "
            "`False` (not erroring out).")
    category_group.add_argument("-cpn","--categories_page_name", default="",
        help="the name of the `categories` page, which will have a list of all"
            " categories and their posts. If left empty, no such page will be created."
            " Default is ''.")
    category_group.add_argument("-gc","--group_categories",
        action="store_const", const=True, default=False,
        help="whether to group category pages under a folder. The folder name will"
            " be 'categories' if nothing is specified in `categories_page_name`")
    category_group.add_argument("-un","--uncategorized_name",
        default="Uncategorized", help="the name for the page containing all"
            " posts which don't have a `category` key in their front matter."
            " Default is 'Uncategorized'.")

    archive_group = parser.add_argument_group("Archive pages",
            "Building an aggregation of posts based on the date they were posted")
    archive_group.add_argument("-abm","--build_archive_by_month",
        action="store_const", const=True, default=False,
        help="whether to build an archive by month. Defaults to `False`.")
    archive_group.add_argument("-aby","--build_archive_by_year",
        action="store_const", const=True, default=False,
        help="whether to build an archive by year. Defaults to `False`.")
    archive_group.add_argument("-apn", "--archive_page_name", default="",
        help="the name of the `archive` page, which will have a list of all"
            " posts by their month and/or year of publishing. If empty, no"
            " such page will be created. Default is ''.")
    archive_group.add_argument("-ga","--group_archive",
        action="store_const", const=True, default=False,
        help="whether to group archive pages under a folder. The folder name will"
            " be 'archive' if nothing is specified in `archive_page_name`")
    archive_group.add_argument("-damp","--display_archive_by_month_in_list_page",
        action="store_const", const=True, default=False,
        help="whether to display the monthly archive in the archive list page."
            " Default is False.")
    archive_group.add_argument("-dayp","--display_archive_by_year_in_list_page",
        action="store_const", const=True, default=False,
        help="whether to display the yearly archive in the archive list page."
            " Default is False.")
    archive_group.add_argument("-amf","--archive_month_format",
        default="%B, %Y",
        help="the `datetime` fomatting to be used for the archive by month."
            " Default is '%%B, %%Y', e.g. 'April, 2022'.")
    archive_group.add_argument("-ayf","--archive_year_format",
        default="%Y",
        help="the `datetime` fomatting to be used for the archive by year."
            " Default is '%%Y', e.g. '2022'.")

    theme_group = parser.add_argument_group("Theme options",
        "Theme may require different options, such as 'dark' or 'light mode, "
        "including a sidebar or not, social media profile links, etc."
        "Since they may be anything the theme developer is responsible for "
        " outlining what arguments may be set. In order for them to be parsed"
        " correctly the following 3 rules must be considered:\n"
        " \n1. flags with no values start with `-th_` or `--theme_option_`,"
        " e.g. `-th_sidebar`.\n"
        " \n2. we support "
        " key-value pairs with up to three values per key. Simple theme args "
        " must start with `-th1_` or `--theme_option1`, e.g. `-th1_dark_mode` or"
        " `--theme_option1_dark_mode`. Arguments that accept tuples with 2 values"
        " should be prefixed with `-th2_` or `--theme_option2` and the same goes"
        " for arguments that accept tuples with 3 values.\n"
        " \n3. Setting the same"
        " argument twice using `-th1_{arg}` will overwrite the previous value."
        " so if a list of values is required use `-thl1_{arg}`,`-thl2_{arg}`,etc."
        "\n\nNOTE: we only support up to 3 values per key.\n\n"
        "Here are the accepted arguments for the default theme:\n"
        "\n\n    th_breadcrumbs - whether to display breadcrumbs or not. Default is off."
        "\n\n    th_sidebar - whether to build a sidebar or not. Default is off."
        "\n\n    th_sidebar_in_home_only - whether to have the sidebar only on the home page."
        "\n\n    th_sidebar_image_in_home_only - whether to have the sidebar "
        " image only on the home page. Default is off."
        "\n\n    th_sidebar_social_in_home_only - whether to have the sidebar "
        " social icons only on the home page. Default is off."
        "\n\n    th_sidebar_description_in_home_only - whether to have the sidebar "
        " description only on the home page. Default is off."
        "\n\n    th1_sidebar_image - an image for the top of the sidebar, e.g. "
            "`-th1_sidebar_image \"/img/sidebar.png\"`"
        "\n\n    th1_sidebar_image_alt - the alt text for the sidebar image"
        "\n\n    th1_sidebar_description - a bit of descriptive text in the sidebar"
        "\n\n    thl{2/3}_social_links - a list of (name, url, optional img) tuples, e.g."
        "`-thl3_social_links github \"https://github.com/lamya\" \"/img/github.svg\""
        " -thl2_social_links twitter \"https://twitter.com/lamya\"`, which will"
        " produce an icon for github and the text 'twitter' for the twitter link."
        "\n\n    thl{2/3}_links - a list of (name, url, optional nice name) tuples, e.g."
        "`-thl3_links \"my knitting blog\" \"https://knittingforfunandprofit.com\""
        " knittingforfunandprofit.com -thl2_links \"favourite knitting pattern\""
        " \"https://knittingforfunandprofit.com/the-three-crocheters\""
        "\n\n    th_sidebar_archive_nav - whether to include an archive navigation"
        " in the sidebar"
        "\n\n    th_display_archive_by_month_in_sidebar - whether to include"
        " monthly links from the archive in the sidebar. Default is True."
        "\n\n    th_display_archive_by_year_in_sidebar - whether to include"
        " yearly links from the archive in the sidebar. Default is False."
        "\n\n    th1_copyright_year - to be displayed in the footer"
        "\n\n    thl1_extra_css - a number of stylesheets to load after"
        " the theme's ones"
        "\n\n    thl1_extra_js - a number of javascript files to load after"
        " the theme's ones"
        )

    return parser

def parse_args():
    parser = build_parser()

    parsed_args, unknown_args = parser.parse_known_args()

    return parsed_args, unknown_args, {long: short for short,long in\
        [a.option_strings for a in parser._get_optional_actions()]}


def process_args(parsed_args, unknown_args):
    if parsed_args.url is None:
        raise MissingRequiredLamyaArgumentError("The -url, --url argument is required.")

    if parsed_args.name is None:
        parsed_args.name = Path(os.getcwd()).stem

    parsed_args.site_directory = Path(parsed_args.site_directory).absolute()
    for _dir in ["content","templates","static","build"]:
        if getattr(parsed_args, _dir + "_directory") is None:
            setattr(parsed_args, _dir + "_directory", parsed_args.site_directory / _dir)

    # Theme options
    theme_options = {}
    simple_option_prefixes = ["-th1_","-th2_","-th3_","--theme_option1_",
         "--theme_option2_","--theme_option3_"]
    list_option_prefixes = ["-thl1_","-thl2_","-thl3_","--theme_list_option1_",
         "--theme_list_option2_","--theme_list_option3_"]
    while unknown_args:
        current = unknown_args.pop(0)
        is_simple_theme_option = any([current.startswith(x) for x in simple_option_prefixes])
        is_list_theme_option = any([current.startswith(x) for x in list_option_prefixes])

        if current.startswith("-th_") or current.startswith("--theme_option_"):
            theme_options[current.replace("-th_","").replace("--theme_option_","")] = True
            continue

        if not (is_simple_theme_option or is_list_theme_option):
            raise UnrecognizedLamyaArgumentError("Unrecognized argument " + current)

        tuple_size = int(current.split("_")[2 if current.startswith("--") else 0][-1])

        arg_values = []
        for i in range(tuple_size):
            if not unknown_args or unknown_args[0].startswith("-"):
                raise NotEnoughValuesForLamyaArgumentError(
                    "Not enough values for arg " + current)
            arg_values.append(unknown_args.pop(0))

        key = current
        for each in simple_option_prefixes + list_option_prefixes:
            key = key.replace(each,"")

        value = tuple(arg_values) if len(arg_values) > 1 else arg_values[0]

        if is_simple_theme_option:
            theme_options[key] = value
        else:
            if not isinstance(theme_options.get(key,[]), list):
                theme_options[key] = []
            theme_options.setdefault(key,[]).append(value)

    if hasattr(parsed_args,"theme_options"):
        for k,v in theme_options.items():
            parsed_args.theme_options[k] = v
    else:
        parsed_args.theme_options = theme_options

    return parsed_args


if __name__ == "__main__":
    parsed_args,unknown_args,option_strings = parse_args()
    if parsed_args.from_file is None:
        site_generator.SiteGenerator.run_from_config(
            (process_args(parsed_args, unknown_args)))
    else:
        # If we are reading from a file, let's get a dict with all the definitions
        args = {}
        with open(parsed_args.from_file, "r") as f:
            exec(f.read(), {}, args)

        # Then create a simple struct mimicking the parsed_args object
        class ArgsFromFile: pass
        args_object = ArgsFromFile

        # Set all the config info on it
        for k,v in args.items():
            setattr(args_object, k, v)

        # Overwrite any settings if they have been provided via the CLI or
        # if they are missing
        for k,v in parsed_args._get_kwargs():
            if ("--"+k in sys.argv or option_strings["--"+k] in sys.argv)\
                    or not hasattr(args_object, k):
                setattr(args_object, k, v)

        site_generator.SiteGenerator.run_from_config(
            process_args(args_object, unknown_args))



"""This module provides a base Config class defining the default parameters
for running a static site generation process, which can be subclassed to overwrite
only the properties you would like."""
from pathlib import Path


class Config:
    """The base Config class defining the default argument values for a static
    site generation process.

    You will most likely want to overwrite this class to change at least a few
    of the arguments.

    NOTE: The arguments here are the same as those when running from the CLI,
    as this Config is meant to be used to run the whole site generation process,
    rather than just the initialization of the :class:`lamya.site_generator.SiteGenerator`
    class.

    :param name: name of the site
    :param url: url of the site
    :param subtitle: optional subtitle of the site, generally used as a short
        description or motto
    :param author_link: a URL to be placed in the `link rel=author` html tag
    :param language: the language to be placed in the html `lang` property
    :param use_absolute_urls: whether or not to use full absolute URLs instead
        of the default relative ones
    :param posts_per_page: how many posts to have per page. Defaults to -1,
        which means no pagination.
    :param publish_date_key: the key in the front matter to read the publish
        date from
    :param read_date_format: the `datetime` format to expect the publish date in
    :param display_date_format: the `datetime` format to display the publish date in
    :param site_directory: the path to the site folder. If none of the following
        paths - content, theme, etc. are provided they will be folders inside
        this directory by default.
    :param content_directory: the path to the top folder containing the site content
    :param theme_directory: the path to the theme folder
    :param static_directory: the path to a folder containing static files to be
        directly copied over to the `build_directory`
    :param templates_directory: the path to a folder with render templates
    :param build_directory: the path to a folder where to generate the website
    :param home_name_in_navigation: the name under which the home page should
        appear in the navigation
    :param exclude_categories_from_navigation: whether or not to exclude the
        category pages from the navigation if they exist
    :param exclude_archive_from_navigation: whether or not to exclude any
        archive pages from the navigation if they exist
    :param exclude_from_navigation: other :mod:`lamya.content_tree` entities
        to exclude from the navigation specified either by name or path, from the navigation if they exist
    :param exclude_from_navigation: other :mod:`lamya.content_tree` entities
        to exclude from the navigation specified either by name or path
    :param custom_navigation: a custom dict-like object specifying the navigation
    :param locally_aggregate_whitelist: a list of folders that should have their
        children aggregated into their index pages. NOTE: only the whitelist OR
        the blacklist can be specified at a time, but not both.
    :param locally_aggregate_blacklisk: a list of folders that should NOT have their
        children aggregated into their index pages. NOTE: only the whitelist OR
        the blacklist can be specified at a time, but not both.
    :param globally_aggregate_whitelist: a list of pages or posts to aggregate
        into the home page. NOTE: only the whitelist OR
        the blacklist can be specified at a time, but not both.
    :param globally_aggregate_blacklist: a list of pages or posts to NOT aggregate
        into the home page. NOTE: only the whitelist OR
        the blacklist can be specified at a time, but not both.
    :param build_categories: whether to build category pages
    :param do_not_allow_uncategorized: whether to error out if `build_categories`
        is True and there are posts without categories
    :param categories_page_name: the name of the page containing all the posts
        grouped by their categories. If not provided, no such page will be generated.
    :param group_categories: whether to group the category pages into a folder
        rather than generating them directly under the root
    :param uncategorized_name: the name used to display uncategorized posts
    :param build_archive_by_month: whether to build a monthly archive
    :param build_archive_by_yearly: whether to build a yearly archive
    :param archive_page_name: the name of the page containing all the posts
        grouped by their publish dates. If not provided, no such page will be generated.
    :param group_archive: whether to group the archive pages into a folder
        rather than generating them directly under the root
    :param display_by_month_in_list_page: whether to include the monthly archive
        in the `archive_page_name` page
    :param display_by_year_in_list_page: whether to include the yearly archive
        in the `archive_page_name` page
    :param archive_month_format: the `datetime` representation to group the
        the posts by month by
    :param archive_year_format: the `datetime` representation to group the
        the posts by year by
    :param theme_options: a dict-like object containing specific to the theme arguments
    """
    # main
    name = "unnamed"
    url = "http://localhost:8000"
    subtitle = ""
    author_link = ""
    language = "en"
    use_absolute_urls = False
    posts_per_page= -1
    publish_date_key = "publish_date"
    read_date_format = "%d-%m-%Y %H:%M"
    display_date_format = "%B %-d, %Y"
    # directories
    site_directory = "."
    content_directory = None
    theme_directory = None
    static_directory = None
    templates_directory = None
    build_directory = None
    # navigation
    home_name_in_navigation = None
    exclude_categories_from_navigation = False
    exclude_archive_from_navigation = False
    exclude_from_navigation = None
    custom_navigation = None
    # aggregation
    locally_aggregate_whitelist=None
    locally_aggregate_blacklist=None
    globally_aggregate_whitelist=None
    globally_aggregate_blacklist=None
    # category pages
    build_categories = False
    do_not_allow_uncategorized = False
    categories_page_name = ""
    group_categories = False
    uncategorized_name = "Uncategorized"
    # archive pages
    build_archive_by_month = False
    build_archive_by_year = False
    archive_page_name = ""
    group_archive = False
    display_archive_by_month_in_list_page = False
    display_archive_by_year_in_list_page = False
    archive_month_format = "%B, %Y"
    archive_year_format = "%Y"
    # theme
    theme_options = None

    def __init__(self):
        self.exclude_from_navigation = self.exclude_from_navigation or []
        self.locally_aggregate_whitelist = self.locally_aggregate_whitelist or []
        self.locally_aggregate_blacklist = self.locally_aggregate_blacklist or []
        self.globally_aggregate_whitelist = self.globally_aggregate_whitelist or []
        self.globally_aggregate_blacklist = self.globally_aggregate_blacklist or []
        self.theme_options = self.theme_options or {}

        self.site_directory = Path(self.site_directory)
        self.content_directory = Path(self.content_directory)\
            if self.content_directory else self.site_directory / "content"
        self.theme_directory = Path(self.theme_directory) if self.theme_directory else None
        self.static_directory = Path(self.static_directory)\
            if self.static_directory else self.site_directory / "static"
        self.templates_directory = Path(self.templates_directory)\
            if self.templates_directory else self.site_directory / "templates"
        self.build_directory = Path(self.build_directory)\
            if self.build_directory else self.site_directory / "build"



"""The :mod:`lamya.content_tree` module provides a naive tree data structure
implementation, specifically designed to serve as a base for generating
static sites. The :mod:`lamya.site_generator` module utilises this to provide
a static site generator, but users are also encouraged to write their own
build scripts directly using the :mod:`lamya.content_tree` module.
"""
from pathlib import Path
from copy import deepcopy
from collections import OrderedDict
from math import ceil

try:
    import markdown
except ImportError:
    markdown = None

from lamya.content_processing import split_front_matter


class ContentTree:
    """A tree implementation specific to parsing, modifying and generating
    the content for a static website.

    This is the abstract base class.

    :param name: the name of the node
    :param user_data: a convenience object for storing custom user data
    :param parent: a handle to the parent node
    :param ancestors: a list of all the ancestors starting from the direct parent
    :param path: a `pathlib.Path` object containing the path to this node starting
        from the root
    :param href: a `pathlib.Path` object containing the path to this node,
        corresponding to what's deemed the correct way referencing the path as a URL
    """
    def __init__(self, name, user_data=None):
        """Constructor

        :param name: the name of the node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__()
        self.name = name
        self._parent = None
        self.user_data = user_data if not user_data is None else {}

    @property
    def parent(self):
        """Get the parent of the node."""
        return self._parent

    @parent.setter
    def parent(self, _): # pylint: disable=no-self-use
        raise AttributeError("'parent' can't be set directly. Use the "
            "'ContentTree.parent_to()' function.")

    def parent_to(self, new_parent):
        """Parent this node to the `new_parent`.

        :param new_parent: the new parent to parent this node to"""
        if not hasattr(new_parent, "children"):
            raise LeafChildError("{new_parent.name} does not accept children.")

        if self._parent:
            self._parent.children.remove(self)
        new_parent.children.append(self)
        self._parent = new_parent

    @property
    def ancestors(self):
        """Get a list of ancestors, starting with the direct parent."""
        _ancestors = []
        p = self
        while p.parent:
            _ancestors.append(p.parent)
            p = p.parent
        return _ancestors

    @property
    def path(self):
        """Get the path of this node relative to the root."""
        return (self.parent.path / self.name)\
            if self.parent else Path("/")

    @path.setter
    def path(self, _): # pylint: disable=no-self-use
        raise AttributeError("'path' can't be set directly. It is managed internally.")

    @property
    def href(self):
        """Returns the path of this node relative to the root, but only if this
        node is not an `index_page` to another node, in which case it returns
        the path of that node instead.

        This corresponds to the fact that in the context of static websites
        the `index.html` page of a folder is in a path like
        `{..}/folder/index.html` and not `{..}/folder/folder/index.html`.
        """
        return self.parent.path if getattr(self, "is_index_page", lambda: False)()\
            else self.path

    def pprint(self, level=0):
        """Returns a pretty string displaying information about this node.

        :param level: the indentation level, so we support nesting"""
        return "{indent1}{type}({name})".format( # pylint: disable=consider-using-f-string
            indent1=" "*2*level, type=self.__class__.__name__, name=self.name)

    def __str__(self):
        return self.pprint()
    def __repr__(self):
        return self.pprint()

    @staticmethod
    def from_directory(directory, accepted_file_types=None, post_create_callback=None):
        """Builds a :class:`ContentTree` structure by walking a directory.

        :param directory: the directory to walk, which would usually be the
            content folder of a static website
        :param accepted_file_types: a list of accepted file extensions. Defaults
            to markdown files only
        :param post_create_callback: a function to be called after each node
            of the content tree is created. Defaults to None"""
        post_create_callback = post_create_callback or (lambda _: None)
        root = Root()
        post_create_callback(root)

        def recursive_builder(parent, directory):
            for child_path in sorted(directory.iterdir()):
                if child_path.is_dir():
                    child = recursive_builder(Folder(child_path.name), child_path)
                    parent.add_child(child)

                elif child_path.suffix not in (accepted_file_types\
                        if accepted_file_types is not None else [".md"]):
                    continue

                elif child_path.stem == "index":
                    child = CustomIndexPage(parent, source_path=child_path)
                    parent.index_page = child

                else:
                    child = PageOrPost(child_path.stem, source_path=child_path)
                    parent.add_child(child)

                post_create_callback(child)

            return parent

        recursive_builder(root, directory)

        return root


class Folder(ContentTree):
    """This class represents a subtree, or in the context of a static website
    a group/folder with pages or groups/folders underneath it.

    You can think of it as the `blog` folder which is the parent to all the
    blog posts in a hypothetical personal website.

    :param children: a list of all the direct children of this node
    :param index_page: the index_page node of this folder. If we follow the
        example of a `blog` folder, the index page is most likely a list of
        all the blog posts
    """
    def __init__(self, name, user_data=None):
        """Constructor

        :param name: the name of the node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__(name, user_data)
        self._children = []
        self._index_page = None

    @property
    def children(self):
        """Returns a list of all the direct children of this node."""
        return self._children

    @children.setter
    def children(self, _): # pylint: disable=no-self-use
        raise AttributeError("'children' can't be set directly. Use the "
            "'ContentTree.parent_to()' function.")

    def add_child(self, child):
        """Adds a child to this node and unparents it from its previous parent.

        :param child: the child to be parented to this node"""
        child.parent_to(self)

    @property
    def index_page(self):
        """The index page of this folder"""
        return self._index_page

    @index_page.setter
    def index_page(self, new_index_page):
        """Sets the index page of the this folder.

        :param new_index_page: the new index page to set
        """
        if new_index_page.parent:
            if new_index_page.is_index_page():
                new_index_page.parent.index_page = None
            else:
                new_index_page.parent.children.remove(new_index_page)
        self._index_page = new_index_page
        new_index_page._parent = self # pylint: disable=protected-access

    @property
    def path(self):
        """Get the path of this node relative to the root."""
        return super().path

    def pprint(self, level=0):
        """Returns a pretty string displaying information about this node.

        :param level: the indentation level, so we support nesting"""
        return "{indent1}{type}({name}) {{\n{indent2}{children}\n{indent2}}}".format( # pylint: disable=consider-using-f-string
            indent1=" "*2*level, indent2=" "*2*(level+1), type=self.__class__.__name__,
            name=self.name, children=("\n" + " "*2*(level+1)).join(
                (["%si %s" % (" "*2*level, self.index_page.pprint(0))] # pylint: disable=consider-using-f-string
                    if self.index_page else []) +\
                 [x.pprint(level+1) for x in self.children]))

    def get(self, path):
        """Gets a descendant of this node by a path.

        :param path: a path, relative to this node, to the desired descendant

        :raises LookupError: if the path does not exist under this node
        """
        path = Path(path)
        path_parts = path.parts

        if not path_parts:
            return self

        first_part = path_parts[0]
        if first_part == "/":
            if len(path_parts) == 1:
                return self
            first_part = path_parts[1]
            path_parts = path_parts[1:]

        next_step = None
        for i,child in enumerate(self._children):
            if first_part in [f"{{{i}}}", child.name] + (
                    [f"page{int(child.pagination.page_number)}"]\
                    if isinstance(child, PaginatedAggregatedPage) else []):
                next_step = child
                break

        if next_step:
            if isinstance(next_step, Folder):
                return next_step.get("/".join(path_parts[1:]))
            return next_step
        raise LookupError(f"The path '{path}' does not exist under {self}")

    def apply_func(self, func, include_index_pages=True, recursive=True):
        """Applies a function to all nodes in this subtree.

        :param func: the function to be applied
        :param include_index_pages: whether to apply the function to index pages
            Default is True
        :param recursive: whether to apply the function to all descendants or
            just the direct children. Default is True (all descendants)
        """
        for child in list(self.children) + ([self.index_page] if\
                self.index_page and include_index_pages else []):
            func(child)

            if isinstance(child, Folder) and recursive:
                child.apply_func(func, include_index_pages)

    def to_groups(self, grouping_func, include_index_pages=True, recursive=True,
            filter_func=lambda x: True):
        """Sorts the descendants of this subtree into one or more groups, based
        on the return value of a grouping function.

        :param grouping_func: the function to group nodes by. It accepts a node
            and should return a string specifying the group to put the node in
        :param include_index_pages: whether to apply the function to index pages.
            Default is True
        :param recursive: whether to apply the function to all descendants or
            just the direct children. Default is True (all descendants)
        :param filter_func: optional function to use as a filter
        """
        grouped = {}
        for child in list(self.children) + ([self.index_page] if\
                self.index_page and include_index_pages else []):
            if filter_func(child):
                grouped.setdefault(grouping_func(child), []).append(child)

            if isinstance(child, Folder) and recursive:
                for k,v in child.to_groups(grouping_func, include_index_pages,
                        filter_func=filter_func).items():
                    existing_values = grouped.setdefault(k, [])
                    existing_values += v

        return grouped

    def flat(self, include_index_pages=True):
        """Returns a one dimensional list of all descendants.

        :param include_index_pages: whether to apply the function to index pages.
            Default is True
        """
        flattened = []
        for child in self.children:
            flattened.append(child)

            if isinstance(child, Folder):
                flattened += child.flat(include_index_pages)

        if include_index_pages and self.index_page:
            flattened.append(self.index_page)

        return flattened

    def leaves(self, include_index_pages=True):
        """Returns all the leaf nodes (pages or posts) in this subtree.

        :param include_index_pages: whether to apply the function to index pages.
            Default is True
        """
        return list(filter(lambda x: isinstance(x, AbstractPageOrPost),
            self.flat(include_index_pages)))

    def filter(self, func, include_index_pages=True, recursive=True):
        """Returns a copy of this subtree with nodes filtered by the `func` arg.

        :param func: the filter function
        :param include_index_pages: whether to apply the function to index pages.
            Default is True
        :param recursive: whether to apply the function to all descendants or
            just the direct children. Default is True (all descendants)
        """
        copy = deepcopy(self)
        copy.filter_in_place(func, include_index_pages, recursive)

        return copy

    def filter_in_place(self, func, include_index_pages=True, recursive=True):
        """Filters nodes in this subtree by the `func` arg.

        :param func: the filter function
        :param include_index_pages: whether to apply the function to index pages.
            Default is True
        :param recursive: whether to apply the function to all descendants or
            just the direct children. Default is True (all descendants)
        """
        to_remove = []

        for child in self.children:
            if isinstance(child, Folder):
                if recursive:
                    child.filter_in_place(func, include_index_pages)
                    if not func(child) and not child.children:
                        to_remove.append(child)
                else:
                    if not func(child):
                        to_remove.append(child)
            elif not func(child):
                to_remove.append(child)

        for child in to_remove:
            self.children.remove(child)

        if include_index_pages and self.index_page and not func(self.index_page):
            self._index_page = None

    def group(self, name, entities):
        """Groups the specified `entities` into a new folder inside of this
        folder.

        :param name: name of the new folder
        :param entities: a list of ContentTree entities to group under the
            new parent

        :raises GroupError: if the list of entities is empty
        """
        if not entities:
            raise GroupError("Can't group an empty list of entities")

        group = Folder(name)
        group.parent_to(self)

        for entity in entities:
            entity.parent_to(group)

        return group

    def as_dict(self, map_leaf=lambda x: x, map_folder=lambda x: x):
        """Returns a `dict` representing this subtree, with leaves and folders
        optionally remapped by mapping functions.

        :param map_leaf: the function to remap leaf nodes with
        :param map_folder: the function to remap folder nodes with
        """
        return OrderedDict([
            (p.name,
             {"self": map_folder(p), "children": (p.as_dict(map_leaf,map_folder))}\
                if isinstance(p, Folder) and p.children else\
             map_leaf(p))\
            for p in self.children])


class Root(Folder):
    """The root of a content tree structure. The only difference with a regular
    folder is that the name is hardcoded to "/".
    """
    def __init__(self, user_data=None):
        """Constructor

        :param user_data: a convenience object for storing custom user data
        """
        super().__init__("/", user_data)


class AbstractPageOrPost(ContentTree):
    """This is the base leaf node of the content tree, which in the context
    of a static website represents a post or a page.

    :param source_path: path to an optional corresponding source path
    :param source: the source for this page or post. Either the source or
        the source_path need to be set on anything except for `Aggregated[Groups]Pages`,
        if the source is ever going to be requested, otherwise an error will be raised.
    :param front_matter: if the front matter has been parsed this is a `dict`
        containing the front matter evaluated from the source
    :param raw_content: this is the raw content of the source, meaning not yet
        processed by any markup processor
    :param content: this is the markup processed content
    """
    def __init__(self, name, source_path=None, source=None, user_data=None):
        """Constructor

        :param name: name of this node
        :param source_path: an optional path to a source file
        :param source: optional source of this node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__(name, user_data)
        self._source_path = source_path
        self._source = source

        self._front_matter = None
        self._raw_content = None
        self._content = None

    @property
    def source_path(self):
        """Returns the path to the source of this node"""
        return self._source_path

    @source_path.setter
    def source_path(self, new_source_path):
        """Sets the path to the source of this node."""
        self._source_path = new_source_path
        if self._source:
            warning(f"The 'source_path' has been updated on {self.name}, but"
                " the source has already been set, so the new 'source_path'"
                " will have no effect.")

    @property
    def source(self):
        """Gets the source of this node. If the `source` has been directly set
        or read from the `source_path` then that's returned. Otherwise, an
        attempt is made to read the source from the `source_path` and if that
        fails a :exc:`PageOrPostWithoutSourceError` is raised.

        :raises PageOrPostWithoutSourceError: if no source can be identified"""
        if self._source is not None:
            return self._source
        if self._source_path:
            with open(self._source_path, "r", encoding="utf-8") as f:
                return f.read()
        raise PageOrPostWithoutSourceError(
            f"'{self.name}' has neither 'source' nor 'source_path' defined.")

    @source.setter
    def source(self, new_source):
        """Sets the `source` of this node directly. If the `source_path` is also
        set, a warning is raised, since the `source` takes precedence over that.

        :param new_source: the new source to set
        """
        self._source = new_source
        if self._source_path:
            warning(f"The 'source' has been updated on {self.name}, but the"
                " 'source_path' has already been set, so the newly set 'source'"
                " will override that")

    def render_path(self, build_path):
        """Get the path of this node relative to the root and append it to the
        `build_path` argument and adding `index.html` at the end.

        :param build_path: a `pathlib.Path` object specifying the path to append
            the current path of the node to
        """
        return Path(build_path) / (self.parent.path.relative_to("/")\
                if self.parent and hasattr(self, "is_index_page") and self.is_index_page()\
                    else self.path.relative_to("/")) / "index.html"

    def is_index_page(self):
        """Returns whether or not this page or post is the index page of its parent"""
        return False if not self.parent else self.parent.index_page == self

    @property
    def front_matter(self):
        """If the front matter has already been parsed, returns a `dict` containing
        it, otherwise returns None and prints a warning."""
        if self._front_matter is None:
            warning(f"'front_matter' was requested on {self.name}, but it's not been "
                "read and parsed yet. Use 'parse_front_matter_and_content()' first")
        return self._front_matter

    @property
    def raw_content(self):
        """If the content has been parsed, returns it raw as taken from the source,
        otherwise returns None and prints a warning"""
        if self._raw_content is None:
            warning(f"'raw_content' was requested on {self.name}, but it's not been "
                "read and parsed yet. Use 'parse_front_matter_and_content()' first")
        return self._raw_content

    @property
    def content(self):
        """If the content has been parsed, returns it processed by the markup
        processor, otherwise returns None and prints a warning."""
        if self._content is None:
            warning(f"'content' was requested on {self.name}, but it's not been "
                "generated yet. Use 'process_content()' first")
        return self._content

    def parse_front_matter_and_content(self,
            front_matter_and_content_split_func = split_front_matter):
        """Takes the source and splits its front matter from its content, storing
        them in `self._front_matter` and `self._raw_content` respectively.

        :param front_matter_and_content_split_func: the function to use for
            splitting the front matter from the content. By default uses the
            :func:`lamya.content_processing.split_front_matter` function.
        """
        self._front_matter, self._raw_content =\
            front_matter_and_content_split_func(self.source)

    def process_content(self,
            markup_processor_func=markdown.markdown if markdown else None):
        """Processes the `raw_content` using the `markup_processor_func`.

        :param markup_processor_func: the function to process the `raw_content`
            with in order to produce the final content. By default it attempts
            to use the `markdown.markdown` function if `markdown` is available.
        """
        self._content = markup_processor_func(self._raw_content or "")


class PageOrPost(AbstractPageOrPost):
    """Giving a more explicit name to the class used for specific pages and posts."""


class CustomIndexPage(PageOrPost):
    """The class used for index pages that have been read from an index file,
    rather than procedurally generated."""
    def __init__(self, parent, source_path=None, source=None, user_data=None):
        """Constructor

        :param parent: the parent this page is the index of
        :param source_path: an optional path to a source file
        :param source: optional source of this node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__(
            parent.name, source_path, source, user_data)
        self.parent_to(parent)


class ProceduralPage(AbstractPageOrPost):
    """The base class for procedurally generated content tree pages.

    :param source: the difference with the :attr:`AbstractPageOrPost.source` property
        is that on :class:`ProceduralPage`s it is not required to have a `source`
        or `source_path` defined, while one of them is required for an
        :class:`AbstractPageOrPost`."""
    @property
    def source(self):
        if self._source is not None:
            return self._source
        if self._source_path:
            with open(self._source_path, "r", encoding="utf-8") as f:
                return f.read()
        return None

    @source.setter
    def source(self, new_source):
        """Sets the `source` of this node directly. If the `source_path` is also
        set, a warning is raised, since the `source` takes precedence over that.

        :param new_source: the new source to set
        """
        self._source = new_source
        if self._source_path:
            warning(f"The 'source' has been updated on {self.name}, but the"
                " 'source_path' has already been set, so the newly set 'source'"
                " will override that")


class AggregatedGroupsPage(ProceduralPage):
    """A procedurally generated page containing groups of posts, a great example
    of which is categories and archives.
    """
    def __init__(self, name, aggregated_grouped_posts, source_path=None, # pylint: disable=too-many-arguments
            source=None, user_data=None):
        """Constructor

        :param name: name of this node
        :param aggregated_grouped_posts: a `dict` with the group names for keys and
            the corresponding lists of posts as the values
        :param source_path: an optional path to a source file
        :param source: optional source of this node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__(
            name, source_path, source, user_data)
        self.aggregated_grouped_posts = aggregated_grouped_posts

    def pprint(self, level=0):
        """Returns a pretty string displaying information about this node.

        :param level: the indentation level, so we support nesting"""
        return "%sAggregatedGroupsPage(%s) {%s}" % ( # pylint: disable=consider-using-f-string
            " "*2*level, self.name,
            ", ".join(k + ": " + ",".join(p.name for p in v)\
            for k,v in self.aggregated_grouped_posts.items()))


class AggregatedPage(ProceduralPage):
    """A procedurally generated page containing a list of posts."""
    def __init__(self, name, aggregated_posts, source_path=None, # pylint: disable=too-many-arguments
            source=None, user_data=None):
        """Constructor

        :param name: name of this node
        :param aggregated_posts: a list of posts
        :param source_path: an optional path to a source file
        :param source: optional source of this node
        :param user_data: a convenience object for storing custom user data
        """
        super().__init__(name, source_path, source, user_data)
        self.aggregated_posts = aggregated_posts

    def pprint(self, level=0):
        """Returns a pretty string displaying information about this node.

        :param level: the indentation level, so we support nesting"""
        return "%sAggregatedPage(%s) [%s]" % ( # pylint: disable=consider-using-f-string
            " "*2*level, self.name,
            ",".join(p.name for p in self.aggregated_posts))

    def paginate(self, num_posts_per_page, post_create_callback=lambda _: None):
        """Splits this page into multiple pages.

        :param num_posts_per_page: number of posts per page
        :param post_create_callback: a function to call on each generated page,
            after its creation
        """
        if num_posts_per_page <= 0:
            raise PaginationError("Can't paginate with less than 1 posts per page")

        pages = []
        for i in range(0, len(self.aggregated_posts), num_posts_per_page):
            pages.append(PaginatedAggregatedPage(
                self.name,self.aggregated_posts[i:i+num_posts_per_page],
                Pagination(i/num_posts_per_page+1,
                    ceil(len(self.aggregated_posts) / num_posts_per_page),
                    None, None, prev_page=None if i == 0 else pages[-1]),
                source_path=self.source_path, source=self._source))
            post_create_callback(pages[-1])
            pages[-1]._front_matter = self._front_matter # pylint: disable=protected-access
            pages[-1]._raw_content = self._raw_content # pylint: disable=protected-access
            pages[-1]._content = self._content # pylint: disable=protected-access
            pages[-1].user_data = dict(self.user_data)

            if len(pages) > 1:
                pages[-2].pagination.next_page = pages[-1]

        for p in pages:
            p.pagination.first_page = pages[0]
            p.pagination.last_page = pages[-1]

        if self.is_index_page():
            self.parent.index_page = pages[0]
            for p in pages[1:]:
                p.parent_to(self.parent)
        else:
            for p in pages:
                p.parent_to(self.parent)
            self.parent.children.remove(self)

        return pages


class PaginatedAggregatedPage(AggregatedPage):
    """A paginated version of the `AggregatedPage` containing only a subset of
    all the posts belonging to the `AggregatedPage` this page was derived from."""
    def __init__(self, name, aggregated_posts, pagination, # pylint: disable=too-many-arguments
            source_path=None, source=None, user_data=None):
        """Constructor

        :param name: name of this node
        :param aggregated_posts: a list of posts
        :param pagination: the :class:`Pagination` object managing this page
        :param source_path: an optional path to a source file
        :param source: optional source of this node
        """
        super().__init__(
            name, aggregated_posts, source_path, source, user_data=None)
        self.pagination = pagination

    def pprint(self, level=0):
        """Returns a pretty string displaying information about this node.

        :param level: the indentation level, so we support nesting"""
        return "%sPaginatedAggregatedPage(%s, %i) [%s]" % ( # pylint: disable=consider-using-f-string\
            " "*2*level, self.name, self.pagination.page_number,
            ",".join(p.name for p in self.aggregated_posts))

    def paginate(self, *args):
        """Attempting to paginated a :class:`PaginatedPage` will raise a :exc:`PaginationError`

        :raises PaginationError:
        """
        raise PaginationError(f"The page '{self.name}' is already paginated.")

    @property
    def path(self):
        """Gets the path of this page.

        This is slightly different since we have 2 extra rules:
        - if the index page of the parent is one of the pages belonging to the
        same pagination object this page does, then we return `{parent_path}/page%%i` "
        - if there's only one page we return an unpaginated path
        """
        if self.pagination.first_page.is_index_page():
            return self.parent.path /\
                (f"page{int(self.pagination.page_number)}")
        if self.pagination.max_page_number == 1:
            return self.parent.path / self.name
        return self.parent.path / self.name /\
            (f"page{int(self.pagination.page_number)}")


class Pagination: # pylint: disable=too-few-public-methods
    """A data struct containing all the information for a paginated aggregated
    page.

    We need an object like this, as the paginated pages are separate pages
    which don't store references to each other.

    :param page_number: the page number corresponding to the page this pagination
        instance belongs to
    :param max_page_number: the last page number
    :param first_page: a reference to the first page in the pagination
    :param last_page: a reference to the last page in the pagination
    :param prev_page: a reference to the prev page in the pagination
    :param next_page: a reference to the next page in the pagination
    """
    def __init__(self, page_number, max_page_number, first_page, last_page, # pylint: disable=too-many-arguments
            prev_page=None, next_page=None):
        self.page_number = page_number
        self.max_page_number = max_page_number
        self.first_page = first_page
        self.last_page = last_page
        self.prev_page = prev_page
        self.next_page = next_page

    def as_navigation_dict(self):
        """Returns a representation of this pagination object as a dictionary
        containing all the information necessary to correctly display a navigation
        structure inside an html template.
        """
        root = str(self.last_page.href).replace(
            f"page{(self.max_page_number)}", "").rstrip("/")
        return {
            "page_number" : self.page_number,
            "max_page_number" : self.max_page_number,
            "first_page_href" : str(self.first_page.href),
            "last_page_href" : str(self.last_page.href),
            "prev_page_href" : str(self.prev_page.href) if self.prev_page else None,
            "next_page_href" : str(self.next_page.href) if self.next_page else None,
            "root": root if root else "/"
        }


class LeafChildError(Exception):
    """An error specifiyng attempted access to the child/subtree of a leaf node"""

class PageOrPostWithoutSourceError(Exception):
    """An error specifiyng attempted access to the source code of a node which
    has none specified neither with a `source_path` nor directly set."""

class PaginationError(Exception):
    """An error when attempting to paginate an `AggregatedPage`, which can be
    triggered due to either not having any posts to paginate OR attempting to
    paginate an already paginated page."""

class GroupError(Exception):
    """An error raised when attempting to group an empty list of items into a subtree."""

def warning(*args):
    """A utility function for displaying warnings.

    The `*args` are directly passed into a print statement.
    """
    print("LamyaWarning:", *args)
